# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Third-party libraries

This module contains packages that were developed outside Pyomo, but are
included in the Pyomo source tree to streamline deployment (e.g., they
are not easily installable via `pip` or `conda`, they are no longer
under active maintenance, or Pyomo requires custom patches).

"""
