# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

__doc__ = """A module providing interface classes to ease use of the data
    structures and modeling utilities provided elsewhere in mpc.

    Code in this module may import from mpc.data and mpc.modeling
    and should not be imported into these other modules.

    """
