# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Pyomo plugin for interacting with NEOS solvers
#

# Static documentation for NEOS solvers
doc = {
    'baron': 'Deterministic global MINLP solver',
    'bonmin': 'Deterministic local MINLP solver',
    'cbc': 'MILP solver',
    'conopt': 'Feasible path NLP solver',
    'couenne': 'Deterministic global MINLP solver',
    'cplex': 'MILP solver',
    'filmint': 'Deterministic local MINLP solver',
    'filter': 'SQP NLP solver',
    'ipopt': 'Interior point NLP solver',
    'knitro': 'Convex NLP/MINLP solver',
    'l-bfgs-b': 'Bound-constrained NLP solver',
    'lancelot': 'Augmented Lagrangian NLP solver',
    'lgo': 'Lipschitz-Continuous Global Optimizer',
    'loqo': 'Interior point NLP solver',
    'minlp': 'Deterministic local MINLP solver',
    'minos': 'NLP solver',
    'minto': 'MILP solver',
    'mosek': 'Interior point NLP solver',
    # 'octeract': 'Deterministic global MINLP solver',
    'ooqp': 'Convex QP solver',
    'path': 'Nonlinear MCP solver',
    'raposa': 'A Global Solver for Polynomial Programming Problems',
    'snopt': 'SQP NLP solver',
}
